/*=================================================================
 *
 * projectL1Inf.C  MEX function that computes a weighted l1Inf projection of a Matrix A 
 *
 * Inputs:
 *
 * A: the [m X n] matrix to project.
 * w: the m dimensional weight vector; w_i corresponds to the weight of row i.
 * c: L1Inf norm constraint
 *
 * Outputs:
 *
 * B : an [m X n] matrix such that sum_i={1:m} {w_i * max_k |B(i,k)| }= c
 *
 * matlab call:
 * 
 * [B]=projectL1Inf(A,c,w);
 
 *=================================================================*/
/* $Revision: 1$ */
#include <math.h>
#include "mex.h"

/* Input Arguments */

#define	A_IN	prhs[0]
#define	C_IN	prhs[1]
#define W_IN    prhs[2]


/* Output Arguments */

#define	B_OUT	plhs[0]

struct residual_t{
    int rowInd;
    double value;
};
        

int compare(const void *_a, const void *_b) {
 
        double *a, *b;      
        a = (double *) _a;
        b = (double *) _b;              
        if(*a==*b)
            return 0;
        if(*a > *b)
            return -1;
        if(*a < *b)
            return 1;                          
}

int compareR(const void *_a, const void *_b) {
 
        struct residual_t *a, *b;      
        a = (struct residual_t *) _a;
        b = (struct residual_t *) _b;
        
        if((*a).value==(*b).value)
            return 0;
        if((*a).value < (*b).value)
            return -1;
        if((*a).value > (*b).value)
            return 1;                          
}


static void computeL1InfProjection(
		   double	B[],
		   double	C,
 		   double	A[],
           double   w[],
           int nRows,
           int nCols
		   )
{
    
    int i, j, index;
    int rowIndex, colIndex, nonZero;
    double *maxes;
    int *sparseMap;
	double *sparseW;
    double maxRow,value,l1InfNorm;
	double *S;
	int indx;
	struct residual_t *R;
	double residual,si; 
	int idxR;
	double Gradient;
	int *Ks;
	double Norm;
	double nextNorm;
	double theta;
	double *mu;
    double sum;
	int r;
    
    maxes = (double*) malloc(sizeof(double)*nRows);
    
    for(i=0;i<nCols*nRows;i++)
        {
              B[i] = A[i];
        }

  
    /* Compute L1Inf norm of A */
    
    l1InfNorm=0;
    nonZero=0;
    for(i=0;i<nRows;i++)
    {
        maxRow=0;
        for (j=0;j<nCols;j++)
        {
            index=j*nRows + i;           
            value=fabs(A[index]);
            if(value>maxRow)
            {
               maxRow=value;
            }
              
         }
         maxes[i]=maxRow;
         l1InfNorm= l1InfNorm + w[i] * maxRow;                  
         
         if(maxRow>0)
         {
            nonZero++;
         }
    }
    
   
    if(l1InfNorm<=C)
    {
        free(maxes);
        return;
    }
   
    /* Compute absolute values of A */ 
        
    sparseMap = (int*) malloc(sizeof(int)*nonZero);
    sparseW= (double*) malloc(sizeof(double)*nonZero);
    index=0;
    
    for(i=0;i<nRows;i++)
    {
       if(maxes[i]>0)
       {
           sparseMap[index]=i;
           sparseW[index]=w[i];
           index++;
       }    
    }
    
  
    S = (double*) malloc(sizeof(double)* (nCols +1) *nonZero);    
    for(i=0;i<(nCols+1)*nonZero;i++)
    {
        S[i]=0;
    }
   
    for(i=0;i<nonZero;i++)
    {
        for (j=0; j<nCols; j++)
        {
           index=j*nRows + sparseMap[i];       
           value=fabs(A[index]);
           index=i*(nCols+1)+j;
           S[index]=value;
        }       
    }  
    
    
    
   
    /* Sort Colums of S */              
    
    indx = 0;        
    for(i=0;i<nonZero;i++)
      { 
         qsort(&S[indx], nCols+1, sizeof(double), &compare);           
         indx=indx + nCols + 1;          
      }
      
  

      /* Compute Residual */
        
      
      R= (struct residual_t*) malloc(sizeof(struct residual_t)*nonZero*(nCols+1));
         

      indx=0; 
      for (i=0;i<nonZero;i++)
      {
          R[indx].rowInd=i;
          R[indx].value=0;          
          residual=S[indx];
          indx++;
          
          for (j=0;j<nCols;j++)
          {
              R[indx].rowInd=i;
              R[indx].value=residual - S[indx] * (j+1);
              residual=residual + S[indx];
              indx++;
          }          
      }
                      

    /* Sort Residual */
      qsort(R, (nCols+1)*nonZero, sizeof(struct residual_t), &compareR);  
       
      
    /* Compute theta */
       
       
       idxR=0;                                          /* points to the current position in R */
       Gradient=0;                                   /* gradient of the N function, at interval [ R[idxR].value , R[idxR+1].value ] */
       Ks = (int*) malloc(sizeof(int)*nonZero);        /* Ks[j] counts how many points of row j we did visit in R, this includes the first point of the current interval;
                                                              -w_j/Ks[j] is the gradient of h_j in the current interval */
       for (i=0;i<nonZero;i++)
       {
           Ks[i]=0;           
       }     
       
       /* to initialize we skip all the entries in R corresponding to 0 residual, for each row i there will be n_i entries, where n_i is the number of maximums of row i */
       while(R[idxR].value<=0.0)
       {
           Ks[R[idxR].rowInd]++;
           idxR++;          
       }
       idxR=idxR-1;
       
       for (i=0;i<nonZero;i++) {
           Gradient += -sparseW[i]/(double) Ks[i];
       }

            
       /* norm at R[idxR].value */
       Norm=l1InfNorm;
       /* norm at next point, i.e. R[idxR+1].value */
       nextNorm = Gradient * (R[idxR+1].value - R[idxR].value ) + Norm;
       
    
       /* keep crossing points in R until we hit the right interval */
       while (nextNorm>C)
        {          
          idxR++;
          Norm=nextNorm;
          r=R[idxR].rowInd;
          Ks[r]++;
          
          /* the r'th gradient becomes 0 when we have crossed nCols+1 points */
          Gradient = 0;
          for (i=0; i<nonZero; ++i) 
          {
             if (Ks[i]<=nCols) 
             {
                Gradient += -sparseW[i]/(double) Ks[i]; 
             }
          } 
          
          nextNorm = Gradient * (R[idxR+1].value - R[idxR].value ) + Norm;
       
       }

       theta = (C - Norm + Gradient*R[idxR].value)/Gradient;
    
       
       /* Compute mu_i for all i */
    
       mu = (double*) malloc(nonZero*sizeof(double));
       sum = 0;
       for (i=0; i<nonZero; ++i) {
          if (Ks[i]<=nCols) {
               j=idxR;
               while (R[j].rowInd!=i) j--;
               mu[i] = S[i*(nCols+1)+Ks[i]-1] - (1/(double) Ks[i])*(theta-R[j].value);              
          }
          else mu[i] = 0;        
          sum += sparseW[i] * mu[i];  
       }
       
       
       /* Create output matrix */
       
      for(i=0;i<nRows*nCols;i++)
      {
           B[i]=A[i];
      }
             

      for(i=0;i<nonZero;i++)
        {
           for(j=0;j<nCols;j++)
           {
              index=(j)*nRows + sparseMap[i]; 
              if(B[index] > mu[i] || -B[index] > mu[i])
              {
                  if(A[index]>=0.0)
                    B[index] = mu[i];
                  else
                    B[index]=-mu[i];  
              }
           }
        }      
                      
      free(maxes);
      free(sparseMap);
      free(S);
      free(R);
      free(Ks);
      free(mu);
      return;   
}

void mexFunction( int nlhs, mxArray *plhs[], 
		  int nrhs, const mxArray*prhs[] )
     
{ 
    double *B; 
    double *A,*c,*w; 
    int nRows,nCols; 
    
    nRows = mxGetM(A_IN); 
    nCols = mxGetN(A_IN);
  
    B_OUT = mxCreateDoubleMatrix(nRows, nCols, mxREAL); 
    if(B_OUT==NULL)
        printf("ERROR: NO MORE HEAP SPACE");
    
    B = mxGetPr(B_OUT);
    A = mxGetPr(A_IN); 
    c = mxGetPr(C_IN);
    w = mxGetPr(W_IN);
    computeL1InfProjection(B,*c,A,w,nRows,nCols); 
    return;
    
}


