function [ grad ] = compute_gradient( X, Sigma, objtype, lambda )
% X: n x p design pool
% Sigma: X'*diag(w)*X, the re-weighted sample covariance
% objtype: {'A', 'D', 'T', 'E', 'I', 'G'}
% lambda: smoothing parameter
% Note: for D-optimality, computing grad for -1/p log det(Sigma)
% return: grad, a 1 x n vector.

    Sigma = Sigma + lambda * eye(size(Sigma,1));

    % if Sigma is close to singular, return inf
    if (cond(Sigma) > 1e10)
        grad = NaN;
        return;
    end 
    
    [n, p] = size(X);
    
    switch objtype
        case {'a', 'A'}
            [U, S] = eig(Sigma, 'vector');
            Lambda = U * diag(S .^ (-1)) * U';
            grad = -1/p * sum((X * Lambda) .^ 2, 2);
        case {'d', 'D'}
            [U, S] = eig(Sigma, 'vector');
            Lambda = U * diag(S .^ (-0.5)) * U';
            grad = -1/p * sum((X * Lambda) .^ 2, 2);
        case {'t', 'T'}
            grad = -p * sum(X .^ 2, 2) / (trace(Sigma)^2);
        case {'e', 'E'}
            [U, S] = eig(Sigma, 'vector');
            [min_sigma, id_sigma] = min(S);
            u = U(:, id_sigma); % p x 1
            grad = -((X*u) .^ 2) / (min_sigma^2);
            %grad = -(X*(Sigma\u)) .^ 2;
        case {'i', 'I'}
            [U, S] = eig(X'*X, 'vector');
            Lambda = U * diag(S .^ 0.5) * U';
            grad = -1/n * sum((X * (Sigma \ Lambda)) .^ 2, 2);
        case {'g', 'G'}
            [maxj, j] = max(diag(X * (Sigma \ X')));
            grad = -(X * (Sigma \ X(j,:)')) .^ 2;
        otherwise
            error(['Unrecognized objtype: ', objtype]);
    end

end

