function [ ret ] = regsel( X, D, k, withrep, param_eps, param_delta, restart )
% Based on paper "Near-Optimal Design of Experiments via Regret
% Minimization" by Allen-Zhu, Li, Singh and Wang, appeared ICML 2017
% Input:
%   - X: n x p design pool
%   - D: n x 1 weight vector (the optimal solution, pi^* in the paper).
%   - k: subset size
%   - withrep: whether with replacement is allowed.
%   - param_eps, param_delta: parameters specified in the paper
%   - restart: whether to use restarting procedure
% Output:
%   - ret: an n x 1 vector, each containing a non-negative integer. sums to
%   1. If withrep=false then each integer is either 0 or 1

    if (param_eps > 0.2 || param_eps <= 0)
        warning('param_eps out of range. reset to 0.1.\n');
        param_eps = 0.2;
    end
    if (param_delta > 0.2 || param_delta <= 0)
        warning('param_delta out of range. reset to 0.1.\n');
        param_delta = 0.2;
    end
    
    % set algorithm parameters: p and alpha
    [n, p] = size(X);
    alpha = p / param_eps;
    eps_search = 1e-3;
    input_X = X;

    % initialization
    ret = zeros(n, 1);
    
    while (sum(ret) < k)
        % whitening
        X = input_X;
        X = X * inv(sqrtm(X'*diag(D)*X));
        
        Lambda = zeros(p, p);
        t = 0;
        c_lower = 0;
        while (sum(ret) < k)
            % stopping criteria
            if (~withrep && restart && rank(Lambda) == p)
                sigma_min = 1.0 / norm(inv(Lambda));
                if (sigma_min > (1-param_eps-4*param_delta) * t / (sum(D) + p/param_eps))
                    break;
                end
            end
            
            % mmwu update
            [c, Z] = find_normalization_constant(Lambda, eps_search, c_lower);
            c_lower = c - eps_search;
            weight_enumerator = sum((X * Z) .^ 2, 2); % n x 1
            weight_denominator = alpha * diag(X * Z * X') + ones(n, 1);
            if (~withrep)
                % remove those points that are already selected
                weight_enumerator = weight_enumerator .* (ret == 0);
            end
            weight = weight_enumerator ./ weight_denominator;
            [max_weight, ind] = max(weight);
            ret(ind) = ret(ind) + 1;
            Lambda = Lambda + X(ind, :)' * X(ind, :);
            
            % remove D(ind) and update t
            t = t + 1;
            if (~withrep)
                D(ind) = 0;
            end
        end
    end
    
end

