function [ ret, lmin_opt ] = regsel_swap( X, D, k, alpha, T, verbose_step )
% Input:
%   - X: n x p design pool
%   - D: n x 1 weight vector (the optimal solution, pi^* in the paper).
%   - k: subset size
%   - alpha: algorithm parameter
%   - T: quit after T iterations of swapping fail to produce a better sol.
% Output:
%   - ret: an n x 1 vector, each containing a non-negative integer.

    if (nargin < 6)
        verbose_step = 1;
    end
    
    if (nargin < 5)
        T = 10;
        verbose_step = 1;
    end

    if (nargin < 4)
        alpha = 10;
        T = 10;
        verbose_step = 1;
    end
    
    D = D / sum(D) * k;
    assert(max(D) <= 1+1e-9);

    % Whitening
    [n, p] = size(X);
    %X = X * inv(sqrtm(X'*diag(D)*X));
	X = X * inv(sqrtm((X .* D(:,ones(1,p)))'*X));
    
    % Initialization
    ret = zeros(n, 1);
    ret(1:k) = 1;
    %Lambda = X'*diag(ret)*X;
	Lambda = X(ret==1,:)'*X(ret==1,:);
    if (rank(Lambda) < p)
        lmin_opt = 0;
    else
        lmin_opt = 1/norm(inv(Lambda));
    end
    C0_opt = ret;
    
    iter = 0;
    iter_progress = 0;
    history = zeros(100, k);
    history(1, :) = sort(find(ret==1));
    while (iter_progress <= T || rank(X(ret==1,:)) < p) 
        
        % compute the At matrix
        %[c, Z] = find_normalization_constant(alpha*(X'*diag(ret)*X), 2e-9, 0);
		[c, Z] = find_normalization_constant(alpha*Lambda, 1e-9, 0);
        sqrtmA = Z;
        A = Z * Z;
		u = sum((X*A).*X, 2);
		v = 2*alpha * sum((X*sqrtmA).*X, 2);
        
        % select an index to remove
        ind = ret .* (v < ones(n, 1)); % n x 1 indicator
        if (sum(ind) == 0)
            break;
        end
        t_ind = find(ind);
        t_u = u(ind==1);
        t_v = v(ind==1);
        [min_val, min_id] = min(t_u ./ (ones(length(t_ind),1) - t_v));
        i = t_ind(min_id);
        
        % select an index to include
        ind = ~ret;
        t_ind = find(ind);
        t_u = u(ind==1);
        t_v = v(ind==1);
        [max_val, max_id] = max(t_u ./ (ones(length(t_ind),1) + t_v));
        j = t_ind(max_id);
        
        % perform the swap and compute the improvements on lambda
        ret(i) = 0;
        ret(j) = 1;
		Lambda = X(ret==1,:)' * X(ret==1,:);
        if (rank(Lambda) < p)
            lmin = 0;
        else
            lmin = 1/norm(inv(Lambda));
        end
        
        if (lmin > lmin_opt)
            lmin_opt = lmin;
            C0_opt = ret;
            iter_progress = 0;
        end
        
        iter = iter + 1;
        iter_progress = iter_progress + 1;
        if (verbose_step > 0 && (mod(iter, verbose_step) == 0))
            fprintf('Iter %d: lmin = %f, lmin_opt = %f\n', iter, lmin, lmin_opt);
        end
        
        t_ret = sort(find(ret == 1)');
        if (max(sum(history(1:iter,:) == t_ret(ones(iter,1),:), 2)) == k)
            break;
        end
        history(iter+1, :) = t_ret;
    end
    
    fprintf('Completed. %d iterations, lmin_opt = %f\n', iter, lmin_opt);
    ret = C0_opt;

end

