function [ w, tot_iters, f_opt ] = solve_pi_emd( X, w0, method, infty_norm, objtype, lambda, T, verbose_step )
% The entropic mirror descent algorithm. See solve_pi_pgd.m for explanation
% of arguments not listed below.
%   -gamma: step size parameter; at epoch t we solve the following opt:
%       argmin_C {gamma_t * <g_t, x> + B_psi(x, x_{t-1})}
%   where gamma_t = gamma / sqrt{t}, g_t = nabla f(x_{t-1}) and B_psi is
%   the Bregman divergence corresponding to entropic constraints
%
%   - method: 1 if the function is differentiable; 0 otherwise
%   resulting in different step size policies


    if (nargin < 8)
        verbose_step = 1;
    end
    if (nargin < 7)
        verbose_step = 1;
        maxT = 1000;
    end
    if (nargin < 6)
        verbose_step = 1;
        maxT = 100;
        lambda = 0;
    end

    [n, p] = size(X);
    assert(infty_norm <= 1);
    assert(infty_norm >= 1/n);
    
    % initialization
    %w = 1/n * ones(n, 1);
    w = w0;
    Sigma = X' * diag(w) * X;
    f_opt = compute_objective(X, Sigma, objtype, lambda);
    w_opt = w;
    if (objtype == 'd' || objtype == 'D'); f_opt = exp(f_opt); end
    fprintf('Start %c-optimality: f_0 = %f\n', objtype, f_opt);
    if (objtype == 'd' || objtype == 'D'); f_opt = log(f_opt); end

    gamma = 1;
    gamma_t = 1;
    grad = compute_gradient(X, Sigma, objtype, lambda); % 1 x n
    w_last = w;
    f_last = f_opt;
    
    % The main iterations
    iter = 0;

    while (iter < T)        
        
        % (unconstrained) entropic mirror descent: careful about numerical
        % issues!
        t_grad = grad - max(grad) * ones(n, 1);
        eps = 1e-9;
        w = w_last .* exp(-gamma_t * t_grad) + eps * ones(n, 1);
        w = w / sum(w);
        
        % project back to the feasible region
        w = project_SimpLinf(w, infty_norm)';
        
        % calculate objectives and quality control
        nSigma = X' * diag(w) * X;
        f_now = compute_objective(X, nSigma, objtype, lambda);
        if (f_now < f_opt)
            f_opt = f_now;
            w_opt = w;
        end
        
        if (method == 1)
            % backtracking line search
            if (f_now - f_last <= 0.5 * grad' * (w - w_last) || gamma_t <= 1e-8)
                f_last = f_now;
                w_last = w;
                grad = compute_gradient(X, nSigma, objtype, lambda);
                gamma_t = 1;
                
                iter = iter + 1;
                if (mod(iter, verbose_step) == 0)
                    if (objtype == 'd' || objtype == 'D'); f_now = exp(f_now); end
                    fprintf('Step %d: objective = %f, best_objective = %f\n', iter, f_now, f_opt);
                end    
            else
                gamma_t = gamma_t * 0.5;
            end
        else
            % diminishing step size
            if (isnan(f_now) || isinf(f_now) || cond(nSigma) > 1e10 || f_now > 2 * f_last)
                gamma = gamma * 0.5;
            else
                w_last = w;
                f_last = f_now;
                grad = compute_gradient(X, nSigma, objtype, lambda);
                
                iter = iter + 1;
                if (mod(iter, verbose_step) == 0)
                    if (objtype == 'd' || objtype == 'D'); f_now = exp(f_now); end
                    fprintf('Step %d: objective = %f, best_objective = %f\n', iter, f_now, f_opt);
                end                 
            end
            %gamma_t = gamma / (iter+1);
            gamma_t = gamma / sqrt(iter+1);
            %gamma_t = gamma;
            %gamma_t = gamma / log(iter+5);
        end

    end
    
    w = w_opt;
    fprintf('Completed. # of nonzero weights = %d\n', sum(w > zeros(n, 1)));
    fprintf('Total # of iterations: %d\n', iter);
    fprintf('Final objective: %f\n', f_opt);
    tot_iters = iter;

end

